@@ -22,12 +22,15 @@
 
 #define PY_SSIZE_T_CLEAN
 
+#ifdef _MSC_VER
+#   include "msvc-stdint.h"
+#else
+#   include <stdint.h>
+#endif
+
 #include <errno.h>
-#include <stdbool.h>
-#include <stdint.h>
 #include <string.h>
 #include <sys/stat.h>
-#include <tgmath.h>
 
 #include "Python.h"
 #include "structmember.h"
@@ -40,7 +43,7 @@
 #include "lmdb.h"
 
 
-// Comment out for copious debug.
+/* Comment out for copious debug. */
 #define NODEBUG
 
 #ifdef NODEBUG
@@ -50,11 +53,9 @@
     "lmdb.cpython: %s:%d: " s "\n", __func__, __LINE__, ## __VA_ARGS__);
 #endif
 
-// Inlining control for compatible compilers.
+/* Inlining control for compatible compilers. */
 #if (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
 #   define NOINLINE __attribute__((noinline))
-#elif defined(_MSC_VER)
-#   define NOINLINE __declspec(noinline)
 #else
 #   define NOINLINE
 #endif
@@ -112,7 +113,7 @@ enum string_id {
     WRITE_S,
     WRITEMAP_S,
 
-    // Must be last.
+    /* Must be last. */
     STRING_ID_COUNT
 };
 
@@ -191,36 +192,17 @@ typedef struct IterObject IterObject;
 typedef struct TransObject TransObject;
 
 
-// --------------------------
-// Buffer object abuse macros
-// --------------------------
+/* ------------------------ */
+/* Python 3.x Compatibility */
+/* ------------------------ */
 
 #if PY_MAJOR_VERSION >= 3
 
-// Python 3.3 kindly exports the struct definitions for us.
 #   define MOD_RETURN(mod) return mod;
 #   define MODINIT_NAME PyInit_cpython
-#   define BUFFER_TYPE PyMemoryViewObject
-#   define MAKE_BUFFER() PyMemoryView_FromMemory("", 0, PyBUF_READ)
-#   define SET_BUFFER(buff, ptr, size) {\
-        (buff)->view.buf = (ptr); \
-        (buff)->view.len = (size); \
-        (buff)->hash = -1; \
-    }
 
 #else
 
-// So evil.
-typedef struct {
-    PyObject_HEAD
-    PyObject *b_base;
-    void *b_ptr;
-    Py_ssize_t b_size;
-    Py_ssize_t b_offset;
-    int b_readonly;
-    long b_hash;
-} PyBufferObject;
-
 #   define PyUnicode_InternFromString PyString_InternFromString
 #   define PyBytes_AS_STRING PyString_AS_STRING
 #   define PyBytes_GET_SIZE PyString_GET_SIZE
@@ -229,18 +211,13 @@ typedef struct {
 #   define _PyBytes_Resize _PyString_Resize
 #   define MOD_RETURN(mod) return
 #   define MODINIT_NAME initcpython
-#   define BUFFER_TYPE PyBufferObject
-#   define MAKE_BUFFER() PyBuffer_FromMemory("", 0)
-#   define SET_BUFFER(buf, ptr, size) {\
-        (buf)->b_hash = -1; \
-        (buf)->b_ptr = (ptr); \
-        (buf)->b_size = (size); \
-    }
+#   define PyMemoryView_FromMemory(x, y, z) PyBuffer_FromMemory(x, y)
+#   define PyBUF_READ 0
 
-// Python 2.5
-#ifndef Py_TYPE
-#   define Py_TYPE(ob) (((PyObject*)(ob))->ob_type)
-#endif
+/* Python 2.5 */
+#   ifndef Py_TYPE
+#       define Py_TYPE(ob) (((PyObject*)(ob))->ob_type)
+#   endif
 
 #endif
 
@@ -287,7 +264,7 @@ struct lmdb_object {
 struct DbObject {
     LmdbObject_HEAD
     /** Python Environment reference. */
-    struct EnvObject *env; // Not refcounted.
+    struct EnvObject *env; /* Not refcounted. */
     /** MDB database handle. */
     MDB_dbi dbi;
 };
@@ -325,8 +302,6 @@ struct TransObject {
     MDB_txn *txn;
     /** Bitfield of trans_flags values. */
     int flags;
-    /** NULL if !TRANS_BUFFERS, or prior to any call to get(). */
-    BUFFER_TYPE *key_buf;
     /** Default database if none specified. */
     DbObject *db;
     /** Number of mutations occurred since start of transaction. Required to
@@ -343,12 +318,6 @@ struct CursorObject {
     int positioned;
     /** MDB-level cursor object. */
     MDB_cursor *curs;
-    /** NULL if trans->buffers==0, or prior to any fetch call. */
-    BUFFER_TYPE *key_buf;
-    /** NULL if trans->buffers==0, or prior to any fetch call. */
-    BUFFER_TYPE *val_buf;
-    /** NULL if trans->buffers==0, or prior to any item() call.*/
-    PyObject *item_tup;
     /** mv_size==0 if positioned==0, otherwise points to current key. */
     MDB_val key;
     /** mv_size==0 if positioned==0, otherwise points to current value. */
@@ -402,23 +371,21 @@ static void link_child(struct lmdb_object *parent, struct lmdb_object *child)
  */
 static void unlink_child(struct lmdb_object *parent, struct lmdb_object *child)
 {
-    if(! parent) {
-        return;
+    if(parent) {
+        struct lmdb_object *prev = child->siblings.prev;
+        struct lmdb_object *next = child->siblings.next;
+        if(prev) {
+            prev->siblings.next = next;
+                 /* If double unlink_child(), this test my legitimately fail: */
+        } else if(parent->children.next == child) {
+            parent->children.next = next;
+        }
+        if(next) {
+            next->siblings.prev = prev;
+        }
+        child->siblings.prev = NULL;
+        child->siblings.next = NULL;
     }
-
-    struct lmdb_object *prev = child->siblings.prev;
-    struct lmdb_object *next = child->siblings.next;
-    if(prev) {
-        prev->siblings.next = next;
-             // If double unlink_child(), this test my legitimately fail:
-    } else if(parent->children.next == child) {
-        parent->children.next = next;
-    }
-    if(next) {
-        next->siblings.prev = prev;
-    }
-    child->siblings.prev = NULL;
-    child->siblings.next = NULL;
 }
 
 #define UNLINK_CHILD(parent, child) unlink_child((void *)parent, (void *)child);
@@ -453,9 +420,9 @@ static void invalidate(struct lmdb_object *parent)
 #define INVALIDATE(parent) invalidate((void *)parent);
 
 
-// ----------
-// Exceptions
-// ----------
+/* ---------- */
+/* Exceptions */
+/* ---------- */
 
 struct error_map {
     int code;
@@ -492,9 +459,9 @@ static const struct error_map error_map[] = {
     {ENOSPC, "DiskError"}
 };
 
-// -------
-// Helpers
-// -------
+/* ------- */
+/* Helpers */
+/* ------- */
 
 /**
  * Describes the type of a struct field.
@@ -549,6 +516,8 @@ dict_from_fields(void *o, const struct dict_field *fields)
     while(fields->type != TYPE_EOF) {
         uint8_t *p = ((uint8_t *) o) + fields->offset;
         unsigned PY_LONG_LONG l = 0;
+        PyObject *lo;
+
         if(fields->type == TYPE_UINT) {
             l = *(unsigned int *)p;
         } else if(fields->type == TYPE_SIZE) {
@@ -557,8 +526,7 @@ dict_from_fields(void *o, const struct dict_field *fields)
             l = (intptr_t) *(void **)p;
         }
 
-        PyObject *lo = PyLong_FromUnsignedLongLong(l);
-        if(! lo) {
+        if(! ((lo = PyLong_FromUnsignedLongLong(l)))) {
             Py_DECREF(dict);
             return NULL;
         }
@@ -574,37 +542,17 @@ dict_from_fields(void *o, const struct dict_field *fields)
     return dict;
 }
 
-/**
- * Given an MDB_val `val`, create a new buffer object to describe it, storing
- * the result in `bufp`. If `*bufp` is not NULL, then it is assumed to already
- * contain a buffer object. In that case simply update the existing object.
- * Return the buffer object on success, or NULL on failure.
- */
-static PyObject * NOINLINE
-buffer_from_val(BUFFER_TYPE **bufp, MDB_val *val)
-{
-    BUFFER_TYPE *buf = *bufp;
-    if(! buf) {
-        buf = (BUFFER_TYPE *) MAKE_BUFFER();
-        if(! buf) {
-            return NULL;
-        }
-        *bufp = buf;
-    }
-
-    SET_BUFFER(buf, val->mv_data, val->mv_size);
-    Py_INCREF(buf);
-    return (PyObject *) buf;
-}
-
 /**
  * Given an MDB_val `val`, convert it to a Python string or bytes object,
  * depending on the Python version. Returns a new reference to the object on
  * sucess, or NULL on failure.
  */
 static PyObject *
-string_from_val(MDB_val *val)
+obj_from_val(MDB_val *val, int as_buffer)
 {
+    if(as_buffer) {
+        return PyMemoryView_FromMemory(val->mv_data, val->mv_size, PyBUF_READ);
+    }
     return PyBytes_FromStringAndSize(val->mv_data, val->mv_size);
 }
 
@@ -640,9 +588,9 @@ val_from_buffer(MDB_val *val, PyObject *buf)
 }
 
 
-// -------------------
-// Concurrency control
-// -------------------
+/* ------------------- */
+/* Concurrency control */
+/* ------------------- */
 
 static NOINLINE PyThreadState *save_thread(void)
 {
@@ -660,11 +608,11 @@ static NOINLINE void restore_thread(PyThreadState *state)
     }
 }
 
-// Like Py_BEGIN_ALLOW_THREADS
+/* Like Py_BEGIN_ALLOW_THREADS */
 #define DROP_GIL \
     { PyThreadState *_save; _save = save_thread();
 
-// Like Py_END_ALLOW_THREADS
+/* Like Py_END_ALLOW_THREADS */
 #define LOCK_GIL \
     restore_thread(_save); }
 
@@ -674,9 +622,9 @@ static NOINLINE void restore_thread(PyThreadState *state)
     LOCK_GIL
 
 
-// ----------
-// Exceptions
-// ----------
+/* ---------- */
+/* Exceptions */
+/* ---------- */
 
 /**
  * Raise an exception appropriate for the given `rc` MDB error code.
@@ -735,9 +683,9 @@ get_fspath(PyObject *src)
 }
 
 
-// ----------------
-// Argument parsing
-// ----------------
+/* ---------------- */
+/* Argument parsing */
+/* ---------------- */
 
 #define OFFSET(k, y) offsetof(struct k, y)
 #define SPECSIZE() (sizeof(argspec) / sizeof(argspec[0]))
@@ -850,13 +798,14 @@ static int NOINLINE
 parse_args(int valid, int specsize, const struct argspec *argspec,
            PyObject *args, PyObject *kwds, void *out)
 {
+    unsigned set = 0;
+    unsigned i;
+
     if(! valid) {
         err_invalid();
         return -1;
     }
 
-    unsigned set = 0;
-    unsigned i;
     if(args) {
         int size = PyTuple_GET_SIZE(args);
         if(size > specsize) {
@@ -918,13 +867,18 @@ db_owner_check(DbObject *db, EnvObject *env)
 }
 
 
-// --------------------------------------------------------
-// Functionality shared between Transaction and Environment
-// --------------------------------------------------------
+/* -------------------------------------------------------- */
+/* Functionality shared between Transaction and Environment */
+/* -------------------------------------------------------- */
 
 static PyObject *
 make_trans(EnvObject *env, DbObject *db, TransObject *parent, int write, int buffers)
 {
+    MDB_txn *parent_txn;
+    TransObject *self;
+    int flags;
+    int rc;
+
     DEBUG("make_trans(env=%p, parent=%p, write=%d, buffers=%d)",
         env, parent, write, buffers)
     if(! env->valid) {
@@ -937,7 +891,7 @@ make_trans(EnvObject *env, DbObject *db, TransObject *parent, int write, int buf
         return NULL;
     }
 
-    MDB_txn *parent_txn = NULL;
+    parent_txn = NULL;
     if(parent) {
         if(parent->flags & TRANS_RDONLY) {
             return err_set("Read-only transactions cannot be nested.", EINVAL);
@@ -953,13 +907,11 @@ make_trans(EnvObject *env, DbObject *db, TransObject *parent, int write, int buf
         return err_set(msg, EACCES);
     }
 
-    TransObject *self = PyObject_New(TransObject, &PyTransaction_Type);
-    if(! self) {
+    if(! ((self = PyObject_New(TransObject, &PyTransaction_Type)))) {
         return NULL;
     }
 
-    int flags = (write && !env->readonly) ? 0 : MDB_RDONLY;
-    int rc;
+    flags = (write && !env->readonly) ? 0 : MDB_RDONLY;
     UNLOCKED(rc, mdb_txn_begin(env->env, parent_txn, flags, &self->txn));
     if(rc) {
         PyObject_Del(self);
@@ -975,7 +927,6 @@ make_trans(EnvObject *env, DbObject *db, TransObject *parent, int write, int buf
 #ifdef HAVE_MEMSINK
     self->sink_head = NULL;
 #endif
-    self->key_buf = NULL;
 
     self->mutations = 0;
     self->flags = 0;
@@ -991,6 +942,9 @@ make_trans(EnvObject *env, DbObject *db, TransObject *parent, int write, int buf
 static PyObject *
 make_cursor(DbObject *db, TransObject *trans)
 {
+    CursorObject *self;
+    int rc;
+
     if(! trans->valid) {
         return err_invalid();
     }
@@ -1000,8 +954,7 @@ make_cursor(DbObject *db, TransObject *trans)
         return NULL;
     }
 
-    CursorObject *self = PyObject_New(CursorObject, &PyCursor_Type);
-    int rc;
+    self = PyObject_New(CursorObject, &PyCursor_Type);
     UNLOCKED(rc, mdb_cursor_open(trans->txn, db->dbi, &self->curs));
     if(rc) {
         PyObject_Del(self);
@@ -1012,11 +965,8 @@ make_cursor(DbObject *db, TransObject *trans)
     OBJECT_INIT(self)
     LINK_CHILD(trans, self)
     self->positioned = 0;
-    self->key_buf = NULL;
-    self->val_buf = NULL;
     self->key.mv_size = 0;
     self->val.mv_size = 0;
-    self->item_tup = NULL;
     self->trans = trans;
     self->last_mutation = trans->mutations;
     Py_INCREF(self->trans);
@@ -1024,9 +974,9 @@ make_cursor(DbObject *db, TransObject *trans)
 }
 
 
-// --------
-// Database
-// --------
+/* -------- */
+/* Database */
+/* -------- */
 
 static DbObject *
 db_from_name(EnvObject *env, MDB_txn *txn, const char *name,
@@ -1034,6 +984,7 @@ db_from_name(EnvObject *env, MDB_txn *txn, const char *name,
 {
     MDB_dbi dbi;
     int rc;
+    DbObject *dbo;
 
     UNLOCKED(rc, mdb_dbi_open(txn, name, flags, &dbi));
     if(rc) {
@@ -1041,14 +992,13 @@ db_from_name(EnvObject *env, MDB_txn *txn, const char *name,
         return NULL;
     }
 
-    DbObject *dbo = PyObject_New(DbObject, &PyDatabase_Type);
-    if(! dbo) {
+    if(! ((dbo = PyObject_New(DbObject, &PyDatabase_Type)))) {
         return NULL;
     }
 
     OBJECT_INIT(dbo)
     LINK_CHILD(env, dbo)
-    dbo->env = env; // no refcount
+    dbo->env = env; /* no refcount */
     dbo->dbi = dbi;
     DEBUG("DbObject '%s' opened at %p", name, dbo)
     return dbo;
@@ -1064,6 +1014,7 @@ txn_db_from_name(EnvObject *env, const char *name,
 {
     int rc;
     MDB_txn *txn;
+    DbObject *dbo;
 
     int begin_flags = (name == NULL || env->readonly) ? MDB_RDONLY : 0;
     UNLOCKED(rc, mdb_txn_begin(env->env, NULL, begin_flags, &txn));
@@ -1072,8 +1023,7 @@ txn_db_from_name(EnvObject *env, const char *name,
         return NULL;
     }
 
-    DbObject *dbo = db_from_name(env, txn, name, flags);
-    if(! dbo) {
+    if(! ((dbo = db_from_name(env, txn, name, flags)))) {
         DROP_GIL
         mdb_txn_abort(txn);
         LOCK_GIL
@@ -1105,6 +1055,10 @@ db_clear(DbObject *self)
 static PyObject *
 db_flags(DbObject *self, PyObject *args, PyObject *kwds)
 {
+    unsigned int f;
+    PyObject *dct;
+    int rc;
+
     struct db_flags {
         TransObject *txn;
     } arg = {NULL};
@@ -1122,15 +1076,12 @@ db_flags(DbObject *self, PyObject *args, PyObject *kwds)
     if(! arg.txn->valid) {
         return err_invalid();
     }
-
-    unsigned int f;
-    int rc = mdb_dbi_flags(arg.txn->txn, self->dbi, &f);
-    if(rc) {
+    if((rc = mdb_dbi_flags(arg.txn->txn, self->dbi, &f))) {
         err_set("mdb_dbi_flags", rc);
         return NULL;
     }
 
-    PyObject *dct = PyDict_New();
+    dct = PyDict_New();
     PyDict_SetItemString(dct, "reverse_key", py_bool(f & MDB_REVERSEKEY));
     PyDict_SetItemString(dct, "dupsort", py_bool(f & MDB_DUPSORT));
     return dct;
@@ -1162,9 +1113,9 @@ static PyTypeObject PyDatabase_Type = {
 };
 
 
-// -----------
-// Environment
-// -----------
+/* ----------- */
+/* Environment */
+/* ----------- */
 
 static int
 env_clear(EnvObject *self)
@@ -1240,6 +1191,13 @@ env_new(PyTypeObject *type, PyObject *args, PyObject *kwds)
         {ARG_SIZE, MAX_SPARE_ITERS_S, OFFSET(env_new, max_spare_iters)}
     };
 
+    PyObject *fspath_obj = NULL;
+    EnvObject *self;
+    const char *fspath;
+    int flags;
+    int rc;
+    int mode;
+
     if(parse_args(1, SPECSIZE(), argspec, args, kwds, &arg)) {
         return NULL;
     }
@@ -1248,9 +1206,7 @@ env_new(PyTypeObject *type, PyObject *args, PyObject *kwds)
         return type_error("'path' argument required");
     }
 
-    PyObject *fspath_obj = NULL;
-    EnvObject *self = PyObject_New(EnvObject, type);
-    if(! self) {
+    if(! ((self = PyObject_New(EnvObject, type)))) {
         return NULL;
     }
 
@@ -1259,7 +1215,6 @@ env_new(PyTypeObject *type, PyObject *args, PyObject *kwds)
     self->main_db = NULL;
     self->env = NULL;
 
-    int rc;
     if((rc = mdb_env_create(&self->env))) {
         err_set("mdb_env_create", rc);
         goto fail;
@@ -1283,7 +1238,7 @@ env_new(PyTypeObject *type, PyObject *args, PyObject *kwds)
     if(! ((fspath_obj = get_fspath(arg.path)))) {
         goto fail;
     }
-    const char *fspath = PyBytes_AS_STRING(fspath_obj);
+    fspath = PyBytes_AS_STRING(fspath_obj);
 
     if(arg.create && arg.subdir && !arg.readonly) {
         struct stat st;
@@ -1297,7 +1252,7 @@ env_new(PyTypeObject *type, PyObject *args, PyObject *kwds)
         }
     }
 
-    int flags = MDB_NOTLS;
+    flags = MDB_NOTLS;
     if(! arg.subdir) {
         flags |= MDB_NOSUBDIR;
     }
@@ -1324,8 +1279,8 @@ env_new(PyTypeObject *type, PyObject *args, PyObject *kwds)
         flags |= MDB_NOMEMINIT;
     }
 
-    // Strip +x.
-    int mode = arg.mode & ~0111;
+    /* Strip +x. */
+    mode = arg.mode & ~0111;
 
     DEBUG("mdb_env_open(%p, '%s', %d, %o);", self->env, fspath, flags, mode)
     UNLOCKED(rc, mdb_env_open(self->env, fspath, flags, mode));
@@ -1408,18 +1363,19 @@ env_copy(EnvObject *self, PyObject *args)
         {ARG_OBJ, PATH_S, OFFSET(env_copy, path)}
     };
 
+    PyObject *fspath_obj;
+    int rc;
+
     if(parse_args(self->valid, SPECSIZE(), argspec, args, NULL, &arg)) {
         return NULL;
     }
     if(! arg.path) {
         return type_error("path argument required");
     }
-    PyObject *fspath_obj = get_fspath(arg.path);
-    if(! fspath_obj) {
+    if(! ((fspath_obj = get_fspath(arg.path)))) {
         return NULL;
     }
 
-    int rc;
     UNLOCKED(rc, mdb_env_copy(self->env, PyBytes_AS_STRING(fspath_obj)));
     Py_CLEAR(fspath_obj);
     if(rc) {
@@ -1492,18 +1448,20 @@ env_info(EnvObject *self)
 static PyObject *
 env_flags(EnvObject *self)
 {
+    PyObject *dct;
+    unsigned int flags;
+    int rc;
+
     if(! self->valid) {
         return err_invalid();
     }
 
-    unsigned int flags;
-    int rc = mdb_env_get_flags(self->env, &flags);
-    if(rc) {
+    if((rc = mdb_env_get_flags(self->env, &flags))) {
         err_set("mdb_env_get_flags", rc);
         return NULL;
     }
 
-    PyObject *dct = PyDict_New();
+    dct = PyDict_New();
     PyDict_SetItemString(dct, "subdir", py_bool(!(flags & MDB_NOSUBDIR)));
     PyDict_SetItemString(dct, "readonly", py_bool(flags & MDB_RDONLY));
     PyDict_SetItemString(dct, "metasync", py_bool(!(flags & MDB_NOMETASYNC)));
@@ -1623,12 +1581,13 @@ static const struct dict_field mdb_stat_fields[] = {
 static PyObject *
 env_stat(EnvObject *self)
 {
+    MDB_stat st;
+    int rc;
+
     if(! self->valid) {
         return err_invalid();
     }
 
-    MDB_stat st;
-    int rc;
     UNLOCKED(rc, mdb_env_stat(self->env, &st));
     if(rc) {
         err_set("mdb_env_stat", rc);
@@ -1645,10 +1604,11 @@ static int env_readers_callback(const char *msg, void *str_)
 {
     PyObject **str = str_;
     PyObject *s = PyUnicode_FromString(msg);
+    PyObject *new;
     if(! s) {
         return -1;
     }
-    PyObject *new = PyUnicode_Concat(*str, s);
+    new = PyUnicode_Concat(*str, s);
     Py_CLEAR(*str);
     *str = new;
     if(! new) {
@@ -1663,12 +1623,12 @@ static int env_readers_callback(const char *msg, void *str_)
 static PyObject *
 env_readers(EnvObject *self)
 {
+    PyObject *str;
     if(! self->valid) {
         return err_invalid();
     }
 
-    PyObject *str = PyUnicode_FromString("");
-    if(! str) {
+    if(! ((str = PyUnicode_FromString("")))) {
         return NULL;
     }
 
@@ -1753,9 +1713,9 @@ static PyTypeObject PyEnvironment_Type = {
 };
 
 
-// -------
-// Cursors
-// -------
+/* ------- */
+/* Cursors */
+/* ------- */
 
 static int
 cursor_clear(CursorObject *self)
@@ -1768,17 +1728,6 @@ cursor_clear(CursorObject *self)
         LOCK_GIL
         self->valid = 0;
     }
-    if(self->key_buf) {
-        SET_BUFFER(self->key_buf, "", 0);
-        Py_CLEAR(self->key_buf);
-    }
-    if(self->val_buf) {
-        SET_BUFFER(self->val_buf, "", 0);
-        Py_CLEAR(self->val_buf);
-    }
-    if(self->item_tup) {
-        Py_CLEAR(self->item_tup);
-    }
     Py_CLEAR(self->trans);
     return 0;
 }
@@ -1969,46 +1918,25 @@ cursor_item(CursorObject *self)
     if(! self->valid) {
         return err_invalid();
     }
-    // Must refresh `key` and `val` following mutation.
+    /* Must refresh `key` and `val` following mutation. */
     if(self->last_mutation != self->trans->mutations &&
        _cursor_get_c(self, MDB_GET_CURRENT)) {
         return NULL;
     }
-    if(self->trans->flags & TRANS_BUFFERS) {
-        if(! buffer_from_val(&self->key_buf, &self->key)) {
-            return NULL;
-        }
-        if(! buffer_from_val(&self->val_buf, &self->val)) {
-            return NULL;
-        }
-        if(! self->item_tup) {
-            self->item_tup = PyTuple_Pack(2, self->key_buf, self->val_buf);
-        }
-        if(! self->item_tup) {
-            return NULL;
-        }
-        Py_INCREF(self->item_tup);
-        return self->item_tup;
-    }
 
-    PyObject *key = string_from_val(&self->key);
-    if(! key) {
-        return NULL;
-    }
-    PyObject *val = string_from_val(&self->val);
-    if(! val) {
-        Py_DECREF(key);
-        return NULL;
-    }
+    int as_buffer = self->trans->flags & TRANS_BUFFERS;
+    PyObject *key = obj_from_val(&self->key, as_buffer);
+    PyObject *val = obj_from_val(&self->val, as_buffer);
     PyObject *tup = PyTuple_New(2);
-    if(! tup) {
-        Py_DECREF(key);
-        Py_DECREF(val);
-        return NULL;
+    if(tup && key && val) {
+        PyTuple_SET_ITEM(tup, 0, key);
+        PyTuple_SET_ITEM(tup, 1, val);
+        return tup;
     }
-    PyTuple_SET_ITEM(tup, 0, key);
-    PyTuple_SET_ITEM(tup, 1, val);
-    return tup;
+    Py_CLEAR(key);
+    Py_CLEAR(val);
+    Py_CLEAR(tup);
+    return NULL;
 }
 
 /**
@@ -2020,19 +1948,12 @@ cursor_key(CursorObject *self)
     if(! self->valid) {
         return err_invalid();
     }
-    // Must refresh `key` and `val` following mutation.
+    /* Must refresh `key` and `val` following mutation. */
     if(self->last_mutation != self->trans->mutations &&
        _cursor_get_c(self, MDB_GET_CURRENT)) {
         return NULL;
     }
-    if(self->trans->flags & TRANS_BUFFERS) {
-        if(! buffer_from_val(&self->key_buf, &self->key)) {
-            return NULL;
-        }
-        Py_INCREF(self->key_buf);
-        return (PyObject *) self->key_buf;
-    }
-    return string_from_val(&self->key);
+    return obj_from_val(&self->key, self->trans->flags & TRANS_BUFFERS);
 }
 
 /**
@@ -2152,7 +2073,7 @@ cursor_replace(CursorObject *self, PyObject *args, PyObject *kwds)
         return err_set("mdb_put", rc);
     }
 
-    PyObject *old = string_from_val(&arg.val);
+    PyObject *old = obj_from_val(&arg.val, 0);
     if(! old) {
         return NULL;
     }
@@ -2191,7 +2112,7 @@ cursor_pop(CursorObject *self, PyObject *args, PyObject *kwds)
     if(! self->positioned) {
         Py_RETURN_NONE;
     }
-    PyObject *old = string_from_val(&self->val);
+    PyObject *old = obj_from_val(&self->val, 0);
     if(! old) {
         return NULL;
     }
@@ -2248,19 +2169,12 @@ cursor_value(CursorObject *self)
     if(! self->valid) {
         return err_invalid();
     }
-    // Must refresh `key` and `val` following mutation.
+    /* Must refresh `key` and `val` following mutation. */
     if(self->last_mutation != self->trans->mutations &&
        _cursor_get_c(self, MDB_GET_CURRENT)) {
         return NULL;
     }
-    if(self->trans->flags & TRANS_BUFFERS) {
-        if(! buffer_from_val(&self->val_buf, &self->val)) {
-            return NULL;
-        }
-        Py_INCREF(self->val_buf);
-        return (PyObject *) self->val_buf;
-    }
-    return string_from_val(&self->val);
+    return obj_from_val(&self->val, self->trans->flags & TRANS_BUFFERS);
 }
 
 static PyObject *
@@ -2417,9 +2331,9 @@ static PyTypeObject PyCursor_Type = {
 };
 
 
-// ---------
-// Iterators
-// ---------
+/* --------- */
+/* Iterators */
+/* --------- */
 
 
 /**
@@ -2485,9 +2399,9 @@ static PyTypeObject PyIterator_Type = {
 };
 
 
-// ------------
-// Transactions
-// ------------
+/* ------------ */
+/* Transactions */
+/* ------------ */
 
 static int
 trans_clear(TransObject *self)
@@ -2729,10 +2643,7 @@ trans_get(TransObject *self, PyObject *args, PyObject *kwds)
         }
         return err_set("mdb_get", rc);
     }
-    if(self->flags & TRANS_BUFFERS) {
-        return buffer_from_val(&self->key_buf, &val);
-    }
-    return string_from_val(&val);
+    return obj_from_val(&val, self->flags & TRANS_BUFFERS);
 }
 
 /**
@@ -2843,7 +2754,7 @@ trans_replace(TransObject *self, PyObject *args, PyObject *kwds)
         return err_set("mdb_put", rc);
     }
 
-    PyObject *old = string_from_val(&arg.value);
+    PyObject *old = obj_from_val(&arg.value, 0);
     if(! old) {
         Py_DECREF((PyObject *) cursor);
         return NULL;
@@ -2898,7 +2809,7 @@ trans_pop(TransObject *self, PyObject *args, PyObject *kwds)
         Py_DECREF((PyObject *)cursor);
         Py_RETURN_NONE;
     }
-    PyObject *old = string_from_val(&cursor->val);
+    PyObject *old = obj_from_val(&cursor->val, 0);
     if(! old) {
         Py_DECREF((PyObject *)cursor);
         return NULL;
